 /**
  * XMMS client lib mainloop / poll() implementation for Mozilla.
  *
  * CRITICAL BUG: nothing sent for about 10s
  * CRITICAL BUG: lockups after 5-20s
  *
  * xmms client lib does its own I/O (networking), via a FD, and so needs to be notified
  * when there's new data in the incoming queue, and get notified when the outgoing queue is ready to
  * accept (more) data.
  *
  * Do to that, we'll create a new thread, and do our own poll()/select() there.
  * We'll do essentially the same as the other mainloops like mainloop.cpp, just that we create a thread
  * and proxy back to main thread to do the read/write by XMMS client lib.
  *
  */

#include "xmmsMainloopMoz.h"
#include "nsServiceManagerUtils.h"
#include "nsThreadUtils.h"
#include "private/pprio.h"
#include "prinrval.h"
#include <iostream>

static void xmmsMainloopMoz_need_out_cb (int need_out, void *data)
{
    std::cout << "need_out " << need_out << std::endl;
    xmmsMainloopMoz* ml = (xmmsMainloopMoz*) data;
    if (!ml)
        return;

    // TODO lock?
    if (need_out)
        ml->mPollFlags |= PR_POLL_WRITE;
    else
        ml->mPollFlags &= ~PR_POLL_WRITE;
    std::cout << "set poll flags" << std::endl;
}

NS_IMPL_ISUPPORTS1(xmmsMainloopMoz, nsIRunnable)

xmmsMainloopMoz::xmmsMainloopMoz(xmmsc_connection_t *c)
  : Xmms::MainloopInterface::MainloopInterface(c),
    mXMMSConn(c)
{
    mPollFlags = PR_POLL_READ | PR_POLL_EXCEPT;
    mPollTimeout = 300;
}

void xmmsMainloopMoz::Init()
{
    int osfd = xmmsc_io_fd_get(mXMMSConn);
    mPRFD = PR_ImportFile(osfd);

    xmmsc_io_need_out_callback_set(mXMMSConn, xmmsMainloopMoz_need_out_cb, this);

    xmmsMainloopMoz_need_out_cb(xmmsc_io_want_out(mXMMSConn), this);
    std::cout << "mainloop inited" << std::endl;
}



class SocketEvent : public nsIRunnable
{
    NS_DECL_ISUPPORTS
public:
    SocketEvent(xmmsc_connection_t* c, xmmsMainloopMoz* ml)
        : mXMMSConn(c), mML(ml) {}
protected:
    xmmsc_connection_t* mXMMSConn;
    xmmsMainloopMoz* mML;
};
NS_IMPL_THREADSAFE_ISUPPORTS1(SocketEvent, nsIRunnable)

class ReadEvent : public SocketEvent
{
public:
    ReadEvent(xmmsc_connection_t* c, xmmsMainloopMoz* ml)
        : SocketEvent(c, ml) {}
    NS_IMETHOD Run()
    {
        std::cout << "ReadEvent in main thread" << std::endl;
        xmmsc_io_in_handle(mXMMSConn);
        std::cout << "Read done" << std::endl;
        return NS_OK;
    }
};

class WriteEvent : public SocketEvent
{
public:
    WriteEvent(xmmsc_connection_t* c, xmmsMainloopMoz* ml)
        : SocketEvent(c, ml) {}
    NS_IMETHOD Run()
    {
        std::cout << "WriteEvent in main thread" << std::endl;
        xmmsc_io_out_handle(mXMMSConn);
        //xmmsSocketHandler_need_out_cb(xmmsc_io_want_out(mXMMSConn), mSH); // TODO done automatically?
        std::cout << "Write done" << std::endl;
        return NS_OK;
    }
};

class DetachedEvent : public SocketEvent
{
public:
    DetachedEvent(xmmsc_connection_t* c, xmmsMainloopMoz* ml)
        : SocketEvent(c, ml) {}
    NS_IMETHOD Run()
    {
        std::cout << "OnSocketDetached in main thread" << std::endl;
        xmmsc_io_disconnect(mXMMSConn);
        std::cout << "OnSocketDetached done" << std::endl;
        return NS_OK;
    }
};

void xmmsMainloopMoz::Close ()
{
    std::cout << "OnSocketDetached in socket thread" << std::endl;
    NS_DispatchToMainThread(new DetachedEvent(mXMMSConn, this));
    std::cout << "OnSocketDetached dispatched" << std::endl;
}

/**
 * Xmms::MainloopInterface.
 * Called by xmmsConnection::connect() / Xmms::Client.setMainloop()
 * on main thread.
 * Starts the new thread.
 */
void xmmsMainloopMoz::run()
{
    NS_NewThread(getter_AddRefs(mThread), this); // TODO event freed by threadmanager? should not
#if 0
    // same as NS_NewThread()
    nsresult rv;
    nsCOMPtr<nsIThreadManager> mgr =
        do_GetService(NS_THREADMANAGER_CONTRACTID);
    if (!mgr) throw new Exception("Could not get threadmanager");

    mgr->NewThread(0, getter_AddRefs(mThread));
    if (!mThread) throw new Exception("Could not create new thread");

    rv = thread->Dispatch(this, NS_DISPATCH_NORMAL);
    if (NS_FAILED(rv)) throw new Exception("Could not dispatch event to new thread");
#endif

    running_ = true;
}

/**
 * nsIRunnable
 * Called (first thing) in background thread.
 * Implements the mainloop doing the poll()
 */
NS_IMETHODIMP xmmsMainloopMoz::Run()
{
    std::cout << "Run() in poll thread" << std::endl;

    while (1) // close?
    {
        PRPollDesc pollList[1];
        pollList[0].fd = mPRFD;
        pollList[0].in_flags = mPollFlags;
        pollList[0].out_flags = 0;

        PRIntervalTime timeout = PR_MillisecondsToInterval(mPollTimeout);
        PRIntervalTime before = PR_IntervalNow();

        PRInt32 n = PR_Poll(pollList, 1, timeout);

        PRUint32 passedMS = PR_IntervalToMilliseconds(PR_IntervalNow() - before);
        std::cerr << "Poll returned after " << passedMS << "ms with n=" << n << std::endl;

        if (n < 0)
        {
            std::cerr << "Poll error " << n << std::endl;
        }
        if (n == 0)
        {
            std::cerr << "Poll timeout?" << std::endl;
        }
        else
        {
            if (pollList[0].out_flags & PR_POLL_READ)
            {
                NS_DispatchToMainThread(new ReadEvent(mXMMSConn, this));
                mPollFlags &= ~PR_POLL_READ;
                std::cout << "ReadReady dispatched" << std::endl;
                // TODO is freed nsIRunnable freed by ThreadManager?
            }
            if (pollList[0].out_flags & PR_POLL_WRITE)
            {
                NS_DispatchToMainThread(new WriteEvent(mXMMSConn, this));
                mPollFlags &= ~PR_POLL_WRITE;
                std::cout << "WriteReady dispatched" << std::endl;
            }
        }
     }
     return NS_OK;
}
